/*-----------------------------------------------3DofDuino 0.02-----------------------------------------------------
-------------------------------Daniel Christiansen - X-Sim->Arduino mega Interface -------------------------------
--------------------------------------daniel.christiansen@gmail.com ----------------------------------------------


--------------------------------------------------USAGE-----------------------------------------------------------
 1) Modify The "Setup" Below. And define our outputs.
 2) Configure X-Sim USO:
   ##protocol## 
   syntax:
   $<value><axis><value><axis><value><axis>#
  
   example:
   $1234X4567Y89ABZ#
  
   1-axis example for x-sim2:
   ~36~~a01~Y~~35~  // Axis in CAPITAL letters
  
   3-axis example for x-sim2:
   ~36~~a01~Y~a02~X~a03~Z~35~ // Axis in CAPITAL letters
    
   Use the USO output settings: 16 bits - HEX - 115200 - 8 - NOPARITY - ONESTOPBIT
   

--------------------------------------------------USAGE-----------------------------------------------------------

*/







//-----------------Includes---------------------
#include <Dans3dofServo.h> //should we only include this if we define servos?
#include <PID_v1.h> //likewise with pid

//--------------Servo Objects-------------------
Servo servo[6];//can we only create the amount we need? For now we create an array of six servos

//--------------PID Objects---------------------
const int numOfPids = 6;
int activePids = 0;
double pidInput[numOfPids];
double pidOutput[numOfPids];
double pidSetpoint[numOfPids];
// Adjust the P-I-D here - referrs to the order of definePwm
PID pid0(&pidInput[0], &pidOutput[0], &pidSetpoint[0],2,5,1, DIRECT);
PID pid1(&pidInput[1], &pidOutput[1], &pidSetpoint[1],2,5,1, DIRECT);
PID pid2(&pidInput[2], &pidOutput[2], &pidSetpoint[2],2,5,1, DIRECT);
PID pid3(&pidInput[3], &pidOutput[3], &pidSetpoint[3],2,5,1, DIRECT);
PID pid4(&pidInput[4], &pidOutput[4], &pidSetpoint[4],2,5,1, DIRECT);
PID pid5(&pidInput[5], &pidOutput[5], &pidSetpoint[5],2,5,1, DIRECT);


//------------Global variables-----------------
boolean debug;
boolean serialHeaderRecieved = false;
char inputString[50];
int numOfConfiguredOutputs = 0;
int numOfServos = 0;
int inputStringIndex=0;

unsigned int output[20][12];



//--------------Setup---------------------
//Change to match your setup

void setup() {
  Serial.begin(115200); //set the serial rate
  debug = false; 
  defineServo('X',9,0,255,32768); 
  definePwm('Y',45,0,1023,32768,50,51,0,52,20000);

}


  /* Usage 
  Please declare the same number of axis as those configured in x-sim
 
  defineServo('<axis from x-sim>',<pinnumber>,<rangeLow>,<rangeHigh>,<startpos>)   
  
  definepwm('<axis from x-sim>',<pwmOutputPin>,<rangeLow>,<rangeHigh>,<startpos>,<LimitSwLowPin>,<LimitSwHighPin><analog PID-feedbackPin><reversePin><pwm frequency)
  
  Be aware that pins <44,45> <6,7,8> <2,3,5> are grouped together - changing the frequency for 
  one pin also changes the frequency for the other pins in that group. 
  <range> is the range of feedback for feedbackPin - typically 0-1023 for the analog input.
  
  "Debug = true" prints a  lot of stuff to the serial port - 
  Use serial monitor and send a corectly formatted package to see debugging messages
  */





void loop() {

//--------------Serial stuff---------------------
  if (Serial.available()) {
    // get the new byte:
    char inChar = (char)Serial.read(); 
    
    
    if (inChar == '$'){ //if the new byte is our header
      inputStringIndex=0; //start at the beginning of our char array
      serialHeaderRecieved = true; //
      if (debug)
        Serial.println("\nHeader found.");
      }
    
    if (inChar == '#'){ //if the new byte is our trailer :)
      serialHeaderRecieved = false; // be ready for the next package
       // Serial.println("\nEnd of Package found.");
        
      }
      
      
      
    if (serialHeaderRecieved == true) { // only begin joining our package when we have recievede a start-of-package
      
      if ((inChar >= 'A' && inChar <= 'F') || (inChar >= '0' && inChar <= '9')){ //we only want HEX digits in our inputstring       
        inputString[inputStringIndex]= inChar; //add the chars to inputstring one at a time as they arrive.
        inputStringIndex ++;
        
        
        if (debug){
          Serial.print("\nHex Digit Found: ");
          Serial.print(inputString[inputStringIndex -1]);
        }
      }
      
      else { //this byte is not a valid HEX caracter - maybe its an axis..

        int inCharAsInt = inChar; //convert from char to integer
        for (int i=0; i < numOfConfiguredOutputs; ++i){ //search the output array for an axis matching our recieved byte
         if (output[i][0] == inCharAsInt){
           if (debug){
             Serial.print("\nOutput Axis found: ");
             Serial.print(output[i][0]);   
           }
           
           inputString[inputStringIndex] = '\0'; //finish our char array
           output[i][5] = strtol(inputString,NULL,16);
           inputStringIndex=0;
           if (debug){
             Serial.print(" With DEC-value: ");
             Serial.print(output[i][5],DEC);
            }
          }
        }
      } 
    }
  }

//--------------Output stuff---------------------

 for (int i=0; i <= numOfConfiguredOutputs; i++){ // run through our outputs,
   
   switch(output[i][2]) {
     
     
     
     
     case 1: // Pwm 

         pidSetpoint[output[i][6]] = map(output[i][5], 0, 65535, output[i][3], output[i][4]); // where we would like to be

         pidInput[output[i][6]] = analogRead(output[i][9]); //where we are 

         pidHandler(output[i][6],'C', NULL); //compute the error offset

        if (pidOutput[output[i][6]] > 0 && pidSetpoint[output[i][6]] > pidInput[output[i][6]])
          {
          digitalWrite(output[i][10],LOW); // set reverse pin low - we are going forward
          analogWrite(output[i][1],pidOutput[output[i][6]]); //send the corrected value to pwm pin
          }
          else if(pidOutput[output[i][6]] < 0 && pidSetpoint[output[i][6]] < pidInput[output[i][6]])
          {
          digitalWrite(output[i][10],HIGH);   // set reverse pin High - we are reversing
          analogWrite(output[i][1],(-1 * pidOutput[output[i][6]])); //send the corrected value to pwm pin 
          }
          else
          {
          analogWrite(output[i][1],0); //motor stopped
          digitalWrite(output[i][10],LOW); // set reverse pin low - we are going forward
        }
                  
         if (debug){     
           Serial.print("\n Input value : ");
           Serial.print(output[i][5],DEC);
           Serial.print("\n scaling to match feedback : ");
           Serial.print(pidSetpoint[output[i][6]],DEC);
           Serial.print("\n feedback : ");
           Serial.print(pidInput[output[i][6]],DEC);
           Serial.print("\n pid correction : ");
           Serial.print(pidOutput[output[i][6]],DEC);
      
         }   
     break;

       
     case 2: // Servo
              

      unsigned int servoval = map(output[i][5], 0, 65535, output[i][3], output[i][4]); //scale it to match the specified motion range 
           if (servoval >= 0 && servoval < 256)
          servo[output[i][6]].write(servoval); 
     break;
      
        
   }
   
   
 }
} // The END













//------------------Functions-----------------------


  void pidHandler(int pidNum, int cmd, int pwmResolution){

      
    switch (pidNum) {
    case 0:
      if (cmd == 'M'){
        pid0.SetMode(AUTOMATIC);
        pid0.SetOutputLimits((-1 * pwmResolution),pwmResolution); 
      }
      if (cmd == 'C')
        pid0.Compute();
      break;
    case 1:
      if (cmd == 'M'){
        pid1.SetMode(AUTOMATIC);
        pid1.SetOutputLimits((-1 * pwmResolution),pwmResolution); 
        }
      if (cmd == 'C')
        pid1.Compute();
      break;  
    case 2:
      if (cmd == 'M'){
        pid2.SetMode(AUTOMATIC);
        pid2.SetOutputLimits((-1 * pwmResolution),pwmResolution); 
        }
      if (cmd == 'C')
        pid2.Compute();
      break;  
    case 3:
      if (cmd == 'M'){
        pid3.SetMode(AUTOMATIC);
        pid3.SetOutputLimits((-1 * pwmResolution),pwmResolution);  
        }
      if (cmd == 'C')
        pid3.Compute();
      break;  
    case 4:
      if (cmd == 'M'){
        pid4.SetMode(AUTOMATIC);
        pid4.SetOutputLimits((-1 * pwmResolution),pwmResolution); 
        }
      if (cmd == 'C')
        pid4.Compute();
      break;  
    case 5:
      if (cmd == 'M'){
        pid5.SetMode(AUTOMATIC);
        pid5.SetOutputLimits((-1 * pwmResolution),pwmResolution);
      } 
      if (cmd == 'C')
        pid5.Compute();
      break;  
  }

  }



void defineServo(int axis, int pinNum, unsigned int rangeLow, unsigned int rangeHigh, unsigned int startPos){
  output[numOfConfiguredOutputs][0] = axis;
  output[numOfConfiguredOutputs][1] = pinNum;
  output[numOfConfiguredOutputs][2] = 2; //servo Type
  output[numOfConfiguredOutputs][3] = rangeLow; 
  output[numOfConfiguredOutputs][4] = rangeHigh;
  output[numOfConfiguredOutputs][5] = startPos;  //is later value
  output[numOfConfiguredOutputs][6] = numOfServos; //servo identifier
    
  servo[numOfServos].attach(pinNum); //set the servo output pin
  numOfConfiguredOutputs ++; //total number of outputs has increased by 1
  numOfServos ++; //so has the total number of servos
  if (debug){
     Serial.print("\nServo ");
     Serial.print(numOfServos);
     Serial.print(" on pin ");
     Serial.print(pinNum);
     Serial.print(" attatched to axis: ");
     char charAxis = axis;
     Serial.print(charAxis);
    }
}

void definePwm(int axis, int pinNum, unsigned int rangeLow, unsigned int rangeHigh, unsigned int startPos, int LimitSwLowPin, int LimitSwHighPin, int feedbackPin, int reversePin, int pwmFrequency){
  output[numOfConfiguredOutputs][0]  = axis;
  output[numOfConfiguredOutputs][1]  = pinNum;
  output[numOfConfiguredOutputs][2]  = 1; //PWM Type
  output[numOfConfiguredOutputs][3]  = rangeLow; //range start  (Analog is at minimum 0) (quadcoder is at minimum 0)
  output[numOfConfiguredOutputs][4]  = rangeHigh; // range stop (analog is max 1023) (quadcoder can be at most 65535)
  output[numOfConfiguredOutputs][5]  = startPos;  //will later be used as value
  output[numOfConfiguredOutputs][6]  = activePids; //our personal pid controller
  output[numOfConfiguredOutputs][7]  = LimitSwLowPin; //inputPin Used to signal end of travel LOW
  output[numOfConfiguredOutputs][8]  = LimitSwHighPin; //input Pin Used to signal end of travel High
  output[numOfConfiguredOutputs][9]  = feedbackPin; // analog input for positioning feedback
  output[numOfConfiguredOutputs][10] = reversePin; //Pin Used to signal reverse
  
  pinMode(output[numOfConfiguredOutputs][1], OUTPUT);
  pinMode(output[numOfConfiguredOutputs][10], OUTPUT); 
  
  pinMode(output[numOfConfiguredOutputs][7], INPUT);
  pinMode(output[numOfConfiguredOutputs][8], INPUT);

 
  
 int Freq = pwmFreq(pwmFrequency,pinNum); //set frequency for pwm by manipulationg the correct timer -- returns the pwm resolution for this pin 
  pidHandler(activePids,'M',Freq); //set pid mode and activate current pid
 
  
  numOfConfiguredOutputs ++; //total number of outputs has increased by 1
  activePids ++; //so has the total number of Pids
  

      
}




int pwmFreq(int desiredFreq, int pinNum) {
    int regCounter = 8000000 / desiredFreq; //cpu freq div by 2 = 8000000 / desired frequency = pwm resolution

  if (pinNum == 2 || pinNum == 3 || pinNum == 5){ //timer 3
    TCCR3A = B00101001; // Phase and frequency correct PWM change at OCRA
    TCCR3B = B10001;  // System clock
    OCR3A  = regCounter; // this is now our pwm resolution
   }  

  if (pinNum == 6 || pinNum == 7 || pinNum == 8){ //timer 4
    TCCR4A = B00101001; // Phase and frequency correct PWM change at OCRA
    TCCR4B = B10001;  // System clock
    OCR4A  = regCounter; // this is now our pwm resolution
   }  

  if (pinNum == 44 || pinNum == 45){ //timer 5 (I didn't know that was here :-) cool huh
    TCCR5A = B00101001; // Phase and frequency correct PWM change at OCRA
    TCCR5B = B10001;  // System clock
    OCR5A  = regCounter; // this is now our pwm resolution
   }



return regCounter; // we will need tihs to scale our output to the pwm resolution
}

